import numpy
import sklearn.ensemble

try:
    from typing import OrderedDict
except ImportError:
    from typing import MutableMapping as OrderedDict


class RandomForestClassifier:
    def __init__(self, **kwargs):
        self.clf = sklearn.ensemble.RandomForestClassifier(**kwargs)

    def __call__(self, raw: numpy.ndarray):
        assert isinstance(raw, numpy.ndarray)
        assert len(raw.shape) == 2
        assert raw.shape[1] == 1
        return self.clf.predict(raw)[:, None]

    def fit(self, raw: numpy.ndarray, target: numpy.ndarray):
        assert isinstance(raw, numpy.ndarray)
        assert len(raw.shape) == 2, raw.shape
        assert isinstance(target, numpy.ndarray)
        assert len(target.shape) == 2, target.shape
        self.clf.fit(raw, target)



from marshmallow import Schema, ValidationError, validates_schema

from bioimageio.spec.shared import fields


class PyBioSchema(Schema):
    bioimageio_description: str = ""


class CiteEntry(PyBioSchema):
    text = fields.String(required=True)
    doi = fields.String(missing=None)
    url = fields.String(missing=None)

    @validates_schema
    def doi_or_url(self, data, **kwargs):
        if data["doi"] is None and data["url"] is None:
            raise ValidationError("doi or url needs to be specified in a citation")


class BaseSpec(PyBioSchema):
    name = fields.String(required=True)
    format_version = fields.String(required=True)
    description = fields.String(required=True)
    cite = fields.Nested(CiteEntry(many=True), required=True)
    authors = fields.List(fields.String(required=True))
    documentation = fields.Path(required=True)
    tags = fields.List(fields.String, required=True)
    license = fields.String(required=True)

    language = fields.String(required=True)
    framework = fields.String(missing=None)
    source = fields.String(required=True)
    required_kwargs = fields.List(fields.String, missing=list)
    optional_kwargs = fields.Dict(fields.String, missing=dict)

    test_input = fields.Path(missing=None)
    test_output = fields.Path(missing=None)
    covers = fields.List(fields.Path, missing=list)


class SpecWithKwargs(PyBioSchema):
    spec: fields.SpecURI
    kwargs = fields.Dict(missing=dict)


class InputShape(PyBioSchema):
    min = fields.List(fields.Integer, required=True)
    step = fields.List(fields.Integer, required=True)

    @validates_schema
    def matching_lengths(self, data, **kwargs):
        min_ = data["min"]
        step = data["step"]
        if min_ is None or step is None:
            return

        if len(min_) != len(step):
            raise ValidationError(f"'min' and 'step' have to have the same length! (min: {min_}, step: {step})")


class OutputShape(PyBioSchema):
    reference_input = fields.String(missing=None)
    scale = fields.List(fields.Float, required=True)
    offset = fields.List(fields.Integer, required=True)

    @validates_schema
    def matching_lengths(self, data, **kwargs):
        scale = data["scale"]
        offset = data["offset"]
        if len(scale) != len(offset):
            raise ValidationError(f"scale {scale} has to have same length as offset {offset}!")


class Array(PyBioSchema):
    name = fields.String(required=True)
    axes = fields.Axes(missing=None)
    data_type = fields.String(required=True)
    data_range = fields.Tuple((fields.Float(allow_nan=True), fields.Float(allow_nan=True)))

    shape: fields.Nested


class InputArray(Array):
    shape = fields.Union([fields.ExplicitShape(), fields.Nested(InputShape)], required=True)


class OutputArray(Array):
    shape = fields.Union([fields.ExplicitShape(), fields.Nested(OutputShape)], required=True)
    halo = fields.List(fields.Integer, missing=None)


class TransformationSpec(BaseSpec):
    dependencies = fields.Dependencies(required=True)
    inputs = fields.Nested(InputArray, required=True)
    outputs = fields.Nested(OutputArray, required=True)


class Transformation(SpecWithKwargs):
    spec = fields.SpecURI(TransformationSpec, required=True)


class Weights(PyBioSchema):
    source = fields.URI(required=True)
    hash = fields.Dict()


class Prediction(PyBioSchema):
    weights = fields.Nested(Weights, missing=None)
    dependencies = fields.Dependencies(missing=None)
    preprocess = fields.Nested(Transformation, many=True, missing=list)
    postprocess = fields.Nested(Transformation, many=True, missing=list)


class ReaderSpec(BaseSpec):
    dependencies = fields.Dependencies(missing=None)
    outputs = fields.Nested(OutputArray, required=True)


class Reader(SpecWithKwargs):
    spec = fields.SpecURI(ReaderSpec)
    transformations = fields.List(fields.Nested(Transformation), missing=list)


class SamplerSpec(BaseSpec):
    dependencies = fields.Dependencies(missing=None)
    outputs = fields.Nested(OutputArray, missing=None)


class Sampler(SpecWithKwargs):
    spec = fields.SpecURI(SamplerSpec)
    readers = fields.List(fields.Nested(Reader, required=True), required=True)


class Optimizer(PyBioSchema):
    source = fields.String(required=True)
    required_kwargs = fields.List(fields.String, missing=list)
    optional_kwargs = fields.Dict(fields.String, missing=dict)


class Setup(PyBioSchema):
    samplers = fields.List(fields.Nested(Sampler, required=True), required=True)
    preprocess = fields.Nested(Transformation, many=True, missing=list)
    postprocess = fields.Nested(Transformation, many=True, missing=list)
    losses = fields.Nested(Transformation, many=True, missing=list)
    optimizer = fields.Nested(Optimizer, missing=None)


class Model(BaseSpec):
    prediction = fields.Nested(Prediction)
    inputs = fields.Nested(InputArray, many=True)
    outputs = fields.Nested(OutputArray, many=True)
    training = fields.Dict(missing=None)

    config = fields.Dict(missing=dict)



from bioimageio.spec.shared import yaml


def test_v0_1(rf_config_path_v0_1):
    from bioimageio.spec.v0_1 import schema

    data = yaml.load(rf_config_path_v0_1)
    errors = schema.Model().validate(data)

    assert not errors, errors



import json
from pathlib import Path

from marshmallow_jsonschema import JSONSchema

import bioimageio.spec


def export_json_schema_from_schema(path: Path, schema: bioimageio.spec.schema.SharedBioImageIOSchema):
    with path.open("w") as f:
        json_schema = JSONSchema().dump(schema)
        json.dump(json_schema, f, indent=4, sort_keys=True)


def export_json_schemas(folder: Path, spec=bioimageio.spec):
    if spec == bioimageio.spec:
        format_version_wo_patch = "latest"
    else:
        format_version_wo_patch = spec.__name__.split(".")[-1]

    export_json_schema_from_schema(folder / f"model_spec_{format_version_wo_patch}.json", spec.schema.Model())
    export_json_schema_from_schema(folder / f"rdf_spec_{format_version_wo_patch}.json", spec.schema.RDF())


if __name__ == "__main__":
    dist = Path(__file__).parent / "../dist"
    dist.mkdir(exist_ok=True)

    export_json_schemas(dist)
    export_json_schemas(dist, bioimageio.spec.v0_3)



import os
import pathlib
import warnings
from functools import singledispatch
from typing import Optional, Sequence, Union

from marshmallow import ValidationError, missing

from bioimageio.spec.shared import BIOIMAGEIO_CACHE_PATH, yaml
from bioimageio.spec.shared.common import get_args
from bioimageio.spec.shared.transformers import download_uri_to_local_path, resolve_raw_node_to_node, resolve_uri
from . import nodes, raw_nodes, schema
from .converters import maybe_convert_model


download_uri_node_to_local_path = download_uri_to_local_path


@singledispatch
def load_raw_model(source) -> raw_nodes.Model:
    raise TypeError(source)


@load_raw_model.register
def _(source: dict) -> raw_nodes.Model:
    data = maybe_convert_model(source)
    tree: raw_nodes.Model = schema.Model().load(data)
    return tree


@load_raw_model.register
def _(source: os.PathLike) -> raw_nodes.Model:
    source = pathlib.Path(source)

    if source.suffix not in (".yml", ".yaml"):
        raise ValidationError(f"invalid suffix {source.suffix} for source {source}")
    elif source.suffix == ".yml":
        warnings.warn(
            "suffix '.yml' is not recommended and will raise a ValidationError in the future. Use '.yaml' instead "
            "(https://yaml.org/faq.html)"
        )

    data = yaml.load(source)

    return load_raw_model(data)


@load_raw_model.register
def _(source: str) -> raw_nodes.Model:
    if pathlib.Path(source).exists():
        # assume source is file path
        source = pathlib.Path(source)
    else:
        # assume source is uri
        source = resolve_uri(source, root_path=BIOIMAGEIO_CACHE_PATH)

    return load_raw_model(source)


def load_model(source: Union[os.PathLike, str, dict], root_path: Optional[os.PathLike] = None) -> nodes.Model:
    if root_path is None:
        if isinstance(source, dict):
            raise TypeError("Require root_path if source is dict to resolve relative paths")

        if isinstance(source, str):
            try:  # is source a path as a string?
                if not pathlib.Path(source).exists():  # only assume source is a path if path exists
                    raise FileNotFoundError(source)
            except (TypeError, FileNotFoundError):
                # assume source is uri
                source = resolve_uri(source, root_path=BIOIMAGEIO_CACHE_PATH)

        if isinstance(source, os.PathLike):
            root_path = pathlib.Path(source).parent
        else:
            raise TypeError(f"Expected pathlib.Path, os.PathLike, str or dict, but got {type(source)}")

    root_path = pathlib.Path(root_path)

    return resolve_raw_node_to_node(load_raw_model(source), root_path=root_path, nodes_module=nodes)


def get_nn_instance(
    node: nodes.Model, weight_order: Sequence[nodes.WeightsFormat] = get_args(nodes.WeightsFormat), **kwargs
):
    assert NotImplementedError("weight_order")
    if isinstance(node, nodes.Model):
        if not isinstance(node.source, nodes.ImportedSource):
            raise ValueError(
                f"Encountered unexpected node.source type {type(node.source)}. "
                f"`get_nn_instance` requires _UriNodeTransformer and _SourceNodeTransformer to be applied beforehand."
            )

        joined_kwargs = {} if node.kwargs is missing else dict(node.kwargs)
        joined_kwargs.update(kwargs)
        return node.source(**joined_kwargs)
    else:
        raise TypeError(node)



"""shared raw nodes that shared transformer act on"""
import dataclasses
from dataclasses import dataclass
from pathlib import Path
from typing import List, Union

try:
    from typing import get_args, get_origin
except ImportError:
    from typing_extensions import get_args, get_origin

from .common import get_args
from marshmallow import missing


@dataclass
class Node:
    def __post_init__(self):
        for f in dataclasses.fields(self):
            if getattr(self, f.name) is missing and (
                get_origin(f.type) is not Union or not isinstance(missing, get_args(f.type))
            ):
                raise TypeError(f"{self.__class__}.__init__() missing required argument: '{f.name}'")


@dataclass
class URI(Node):
    """URI as  scheme:[//authority]path[?query][#fragment]"""

    scheme: str = missing
    authority: str = missing
    path: str = missing
    query: str = missing
    fragment: str = missing

    def __str__(self):
        """scheme:[//authority]path[?query][#fragment]"""
        return (
            (self.scheme + ":" if self.scheme else "")
            + ("//" + self.authority if self.authority else "")
            + self.path
            + ("?" + self.query if self.query else "")
            + ("#" + self.fragment if self.fragment else "")
        )


@dataclass
class ImportablePath(Node):
    filepath: Path = missing
    callable_name: str = missing


@dataclass
class ImportableModule(Node):
    module_name: str = missing
    callable_name: str = missing


@dataclass
class ImplicitInputShape(Node):
    min: List[float] = missing
    step: List[float] = missing

    def __len__(self):
        return len(self.min)


@dataclass
class ImplicitOutputShape(Node):
    reference_input: str = missing
    scale: List[float] = missing
    offset: List[int] = missing

    def __len__(self):
        return len(self.scale)



from copy import deepcopy
from pathlib import Path
from pprint import pprint

import requests
import typer
from marshmallow import ValidationError
from ruamel.yaml import YAML

from bioimageio.spec import maybe_convert_model, maybe_convert_manifest, schema

yaml = YAML(typ="safe")

app = typer.Typer()  # https://typer.tiangolo.com/


@app.command()
def package_model(model_yaml: Path, auto_convert: bool = False):
    pass


@app.command()
def verify_spec(model_yaml: Path, auto_convert: bool = False):
    try:
        model_data = yaml.load(model_yaml)
    except Exception as e:
        pprint(e)
        code = 1
    else:
        try:
            if auto_convert:
                model_data = maybe_convert_model(deepcopy(model_data))

            verify_model_data(model_data)
        except ValidationError as e:
            pprint(e.messages)
            code = 1
        else:
            code = 0
            print(f"successfully verified model {model_yaml}")

    raise typer.Exit(code=code)


def verify_model_data(model_data: dict):
    schema.Model().load(model_data)


def verify_bioimageio_manifest_data(manifest_data: dict, auto_convert: bool = False):
    try:
        if auto_convert:
            manifest_data = maybe_convert_manifest(deepcopy(manifest_data))

        manifest = schema.BioImageIoManifest().load(manifest_data)
    except ValidationError as e:
        pprint(e.messages)
        return 1

    code = 0
    for model in manifest["model"]:
        try:
            response = requests.get(model["source"], stream=True)
            model_data = yaml.load(response.content)
            if auto_convert:
                model_data = maybe_convert_model(model_data)

            verify_model_data(model_data)
        except ValidationError as e:
            print("invalid model:", model["source"])
            pprint(e.messages)
            code = 1
        except Exception as e:
            print("invalid model:", model["source"])
            pprint(e)
            code = 1

    return code


@app.command()
def verify_bioimageio_manifest(manifest_yaml: Path, auto_convert: bool = False):
    try:
        manifest_data = yaml.load(manifest_yaml)
    except Exception as e:
        print("invalid manifest", manifest_yaml)
        pprint(e)
        code = 1
    else:
        code = verify_bioimageio_manifest_data(manifest_data, auto_convert=auto_convert)
        if code == 0:
            print(f"successfully verified manifest {manifest_yaml}")

    raise typer.Exit(code=code)


if __name__ == "__main__":
    app()



from pathlib import Path

from bioimageio.spec.shared.nodes import ImportedSource
from .raw_nodes import *


@dataclass
class RDF(RDF):
    covers: List[Path] = missing
    documentation: Path = missing


@dataclass
class WeightsEntry(WeightsEntry):
    source: Path = missing


@dataclass
class Model(Model):
    source: ImportedSource = missing
    test_inputs: List[Path] = missing
    test_outputs: List[Path] = missing
    weights: Dict[WeightsFormat, WeightsEntry] = missing



from bioimageio.spec.shared import fields
from . import nodes, raw_nodes, schema, utils
from .converters import maybe_convert_manifest, maybe_convert_model
from .raw_nodes import ModelFormatVersion

fields = fields


get_nn_instance = utils.get_nn_instance
download_uri_to_local_path = utils.download_uri_to_local_path
load_raw_model = utils.load_raw_model
load_model = utils.load_model



from bioimageio.spec.shared import fields
from . import nodes, raw_nodes, schema
from .raw_nodes import ModelFormatVersion

fields = fields


def maybe_convert_model(data):
    return data


def maybe_convert_manifest(data):
    return data



import dataclasses
import importlib.util
import logging
import pathlib
import sys
import typing
import uuid
import warnings
from functools import singledispatch
from types import ModuleType
from urllib.parse import urlunparse
from urllib.request import url2pathname, urlretrieve

from marshmallow import ValidationError, missing

from . import BIOIMAGEIO_CACHE_PATH, fields, nodes, raw_nodes

GenericNode = typing.TypeVar("GenericNode", bound=raw_nodes.Node)


class Transformer:
    def transform(self, node: GenericNode) -> GenericNode:
        method = "transform_" + node.__class__.__name__

        transformer = getattr(self, method, self.generic_transformer)

        return transformer(node)

    def generic_transformer(self, node: GenericNode) -> GenericNode:
        return node

    def transform_list(self, node: list) -> list:
        return [self.transform(subnode) for subnode in node]

    def transform_dict(self, node: dict) -> dict:
        return {key: self.transform(value) for key, value in node.items()}


def iter_fields(node: dataclasses.dataclass):
    for field in dataclasses.fields(node):
        yield field.name, getattr(node, field.name)


class NodeVisitor:
    def visit(self, node: typing.Any) -> None:
        method = "visit_" + node.__class__.__name__

        visitor = getattr(self, method, self.generic_visit)

        visitor(node)

    def generic_visit(self, node):
        """Called if no explicit visitor function exists for a node."""
        if isinstance(node, raw_nodes.Node):
            for field, value in iter_fields(node):
                self.visit(value)
        elif isinstance(node, list):
            [self.visit(subnode) for subnode in node]
        elif isinstance(node, dict):
            [self.visit(subnode) for subnode in node.values()]
        elif isinstance(node, tuple):
            assert not any(
                isinstance(subnode, raw_nodes.Node) or isinstance(subnode, list) or isinstance(subnode, dict)
                for subnode in node
            )


class NodeTransformer(Transformer):
    def generic_transformer(self, node: GenericNode) -> GenericNode:
        if isinstance(node, raw_nodes.Node):
            return dataclasses.replace(
                node,
                **{field.name: self.transform(getattr(node, field.name)) for field in dataclasses.fields(node)},  # noqa
            )
        else:
            return super().generic_transformer(node)


@dataclasses.dataclass
class LocalImportableModule(raw_nodes.ImportableModule):
    root_path: pathlib.Path = missing


@dataclasses.dataclass
class ResolvedImportablePath(raw_nodes.ImportablePath):
    pass


class UriNodeTransformer(NodeTransformer):
    def __init__(self, *, root_path: pathlib.Path):
        self.root_path = root_path

    def transform_URI(self, node: raw_nodes.URI) -> pathlib.Path:
        local_path = resolve_uri(node, root_path=self.root_path)
        return local_path

    def transform_ImportablePath(self, node: raw_nodes.ImportablePath) -> ResolvedImportablePath:
        return ResolvedImportablePath(
            filepath=(self.root_path / node.filepath).resolve(), callable_name=node.callable_name
        )

    def transform_ImportableModule(self, node: raw_nodes.ImportableModule) -> LocalImportableModule:
        return LocalImportableModule(**dataclasses.asdict(node), root_path=self.root_path)

    def _transform_Path(self, leaf: pathlib.Path):
        assert not leaf.is_absolute()
        return self.root_path / leaf

    def transform_PosixPath(self, leaf: pathlib.PosixPath) -> pathlib.Path:
        return self._transform_Path(leaf)

    def transform_WindowsPath(self, leaf: pathlib.WindowsPath) -> pathlib.Path:
        return self._transform_Path(leaf)


class SourceNodeTransformer(NodeTransformer):
    """
    Imports all source callables
    note: Requires previous transformation by UriNodeTransformer
    """

    class TemporaryInsertionIntoPythonPath:
        def __init__(self, path: str):
            self.path = path

        def __enter__(self):
            sys.path.insert(0, self.path)

        def __exit__(self, exc_type, exc_value, traceback):
            sys.path.remove(self.path)

    def transform_LocalImportableModule(self, node: LocalImportableModule) -> nodes.ImportedSource:
        with self.TemporaryInsertionIntoPythonPath(str(node.root_path)):
            module = importlib.import_module(node.module_name)

        return nodes.ImportedSource(factory=getattr(module, node.callable_name))

    @staticmethod
    def transform_ImportableModule(node):
        raise RuntimeError(
            "Encountered raw_nodes.ImportableModule in _SourceNodeTransformer. Apply _UriNodeTransformer first!"
        )

    @staticmethod
    def transform_ResolvedImportablePath(node: ResolvedImportablePath) -> nodes.ImportedSource:
        importlib_spec = importlib.util.spec_from_file_location(f"user_imports.{uuid.uuid4().hex}", node.filepath)
        dep = importlib.util.module_from_spec(importlib_spec)
        importlib_spec.loader.exec_module(dep)
        return nodes.ImportedSource(factory=getattr(dep, node.callable_name))

    @staticmethod
    def transform_ImportablePath(node):
        raise RuntimeError(
            "Encountered raw_nodes.ImportablePath in _SourceNodeTransformer. Apply _UriNodeTransformer first!"
        )


GenericRawNode = typing.TypeVar("GenericRawNode", bound=raw_nodes.Node)
GenericResolvedNode = typing.TypeVar("GenericResolvedNode", bound=nodes.Node)


class RawNodeTypeTransformer(NodeTransformer):
    def __init__(self, nodes_module: ModuleType):
        super().__init__()
        self.nodes = nodes_module

    def generic_transformer(self, node: GenericRawNode) -> GenericResolvedNode:
        if isinstance(node, raw_nodes.Node):
            resolved_data = {
                field.name: self.transform(getattr(node, field.name)) for field in dataclasses.fields(node)
            }
            resolved_node_type: typing.Type[GenericResolvedNode] = getattr(self.nodes, node.__class__.__name__)
            return resolved_node_type(**resolved_data)
        else:
            return super().generic_transformer(node)


@singledispatch
def resolve_uri(uri, root_path: pathlib.Path = pathlib.Path()) -> pathlib.Path:
    raise TypeError(type(uri))


@resolve_uri.register
def _(uri: raw_nodes.URI, root_path: pathlib.Path = pathlib.Path()) -> pathlib.Path:
    if uri.scheme == "":  # relative path
        if uri.authority or uri.query or uri.fragment:
            raise ValidationError(f"Invalid Path/URI: {uri}")

        local_path = root_path / uri.path
    elif uri.scheme == "file":
        if uri.authority or uri.query or uri.fragment:
            raise NotImplementedError(uri)

        local_path = pathlib.Path(url2pathname(uri.path))
    elif uri.scheme == "https":
        local_path = _download_uri_to_local_path(uri)
    else:
        raise ValueError(f"Unknown uri scheme {uri.scheme}")

    return local_path


@resolve_uri.register
def _(uri: str, root_path: pathlib.Path = pathlib.Path()) -> pathlib.Path:
    return resolve_uri(fields.URI().deserialize(uri), root_path)


@resolve_uri.register
def _(uri: pathlib.Path, root_path: pathlib.Path = pathlib.Path()) -> pathlib.Path:
    return resolve_uri(uri.as_uri(), root_path)


def download_uri_to_local_path(uri: typing.Union[raw_nodes.URI, str]) -> pathlib.Path:
    return resolve_uri(uri)


def _download_uri_to_local_path(uri: typing.Union[nodes.URI, raw_nodes.URI]) -> pathlib.Path:
    local_path = BIOIMAGEIO_CACHE_PATH / uri.scheme / uri.authority / uri.path.strip("/") / uri.query
    if local_path.exists():
        warnings.warn(f"found cached {local_path}. Skipping download of {uri}.")
    else:
        local_path.parent.mkdir(parents=True, exist_ok=True)
        url_str = urlunparse([uri.scheme, uri.authority, uri.path, "", uri.query, uri.fragment])
        try:
            urlretrieve(url_str, str(local_path))
        except Exception:
            logging.getLogger("download").error("Failed to download %s", uri)
            raise

    return local_path


def resolve_raw_node_to_node(
    raw_node: GenericRawNode, root_path: pathlib.Path, nodes_module: ModuleType
) -> GenericNode:
    """resolve all uris and sources"""
    node = UriNodeTransformer(root_path=root_path).transform(raw_node)
    node = SourceNodeTransformer().transform(node)
    node = RawNodeTypeTransformer(nodes_module).transform(node)
    return node



"""shared nodes that shared transformer act on"""

from .raw_nodes import *


@dataclass
class ImportedSource:
    factory: callable

    def __call__(self, *args, **kwargs):
        return self.factory(*args, **kwargs)



import distutils.version
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, NewType, Tuple, Union

from marshmallow import missing
from marshmallow.utils import _Missing

from bioimageio.spec.shared.raw_nodes import (
    ImplicitInputShape,
    ImplicitOutputShape,
    ImportableModule,
    ImportablePath,
    Node,
    URI,
)

from bioimageio.spec.shared.common import Literal, get_args

# Ideally only the current format version is valid.
# Older formats may be converter through `bioimageio.spec.utils.maybe_convert`,
# such that we only need to support the most up-to-date version.
GeneralFormatVersion = Literal["0.2.0"]  # newest format needs to be last (used in spec.__init__.py)
ModelFormatVersion = Literal["0.3.0", "0.3.1", "0.3.2"]  # newest format needs to be last (used in spec.__init__.py)
latest_version = get_args(ModelFormatVersion)[-1]

Axes = NewType("Axes", str)
Dependencies = NewType("Dependencies", str)
Framework = Literal["scikit-learn", "pytorch", "tensorflow"]
Language = Literal["python", "java"]
PreprocessingName = Literal["binarize", "clip", "scale_linear", "sigmoid", "zero_mean_unit_variance", "scale_range"]
PostprocessingName = Literal[
    "binarize", "clip", "scale_linear", "sigmoid", "zero_mean_unit_variance", "scale_range", "scale_mean_variance"
]
Type = Literal["model", "dataset", "application", "notebook"]
WeightsFormat = Literal[
    "pickle",
    "pytorch_state_dict",
    "pytorch_script",
    "keras_hdf5",
    "tensorflow_js",
    "tensorflow_saved_model_bundle",
    "onnx",
]


@dataclass
class Author(Node):
    name: str = missing
    affiliation: Union[_Missing, str] = missing
    orcid: Union[_Missing, str] = missing


@dataclass
class Badge(Node):
    label: str = missing
    icon: Union[_Missing, str] = missing
    url: Union[_Missing, URI] = missing


@dataclass
class CiteEntry(Node):
    text: str = missing
    doi: Union[_Missing, str] = missing
    url: Union[_Missing, str] = missing


ImportableSource = Union[ImportableModule, ImportablePath]


@dataclass
class RunMode(Node):
    name: str = missing
    kwargs: Union[_Missing, Dict[str, Any]] = missing


@dataclass
class RDF(Node):
    attachments: Union[_Missing, Dict[str, Any]] = missing
    authors: List[Author] = missing
    badges: Union[_Missing, List[Badge]] = missing
    cite: List[CiteEntry] = missing
    config: Union[_Missing, dict] = missing
    covers: Union[_Missing, List[URI]] = missing
    description: str = missing
    documentation: Path = missing
    format_version: ModelFormatVersion = missing
    git_repo: Union[_Missing, str] = missing
    license: Union[_Missing, str] = missing
    links: List[str] = missing
    name: str = missing
    tags: List[str] = missing
    type: Type = missing
    version: Union[_Missing, distutils.version.StrictVersion] = missing

    def __post_init__(self):
        if self.type is missing:
            self.type = self.__class__.__name__.lower()  # noqa

        assert self.type in get_args(Type)


@dataclass
class Preprocessing:
    name: PreprocessingName = missing
    kwargs: Dict[str, Any] = missing


@dataclass
class Postprocessing:
    name: PostprocessingName = missing
    kwargs: Dict[str, Any] = missing


@dataclass
class InputTensor:
    name: str = missing
    data_type: str = missing
    axes: Axes = missing
    shape: Union[List[int], ImplicitInputShape] = missing
    preprocessing: Union[_Missing, List[Preprocessing]] = missing
    description: Union[_Missing, str] = missing
    data_range: Union[_Missing, Tuple[float, float]] = missing


@dataclass
class OutputTensor:
    name: str = missing
    data_type: str = missing
    axes: Axes = missing
    shape: Union[List[int], ImplicitOutputShape] = missing
    halo: Union[_Missing, List[int]] = missing
    postprocessing: Union[_Missing, List[Postprocessing]] = missing
    description: Union[_Missing, str] = missing
    data_range: Union[_Missing, Tuple[float, float]] = missing


@dataclass
class WeightsEntry(Node):
    authors: Union[_Missing, List[Author]] = missing
    attachments: Union[_Missing, Dict] = missing
    parent: Union[_Missing, str] = missing
    # ONNX specific
    opset_version: Union[_Missing, int] = missing
    # tag: Optional[str]  # todo: check schema. only valid for tensorflow_saved_model_bundle format
    # todo: check schema. only valid for tensorflow_saved_model_bundle format
    sha256: Union[_Missing, str] = missing
    source: URI = missing
    tensorflow_version: Union[_Missing, distutils.version.StrictVersion] = missing


@dataclass
class ModelParent(Node):
    uri: URI = missing
    sha256: str = missing


@dataclass
class Model(RDF):
    dependencies: Union[_Missing, Dependencies] = missing
    framework: Union[_Missing, Framework] = missing
    inputs: List[InputTensor] = missing
    kwargs: Union[_Missing, Dict[str, Any]] = missing
    language: Union[_Missing, Language] = missing
    license: str = missing
    outputs: List[OutputTensor] = missing
    packaged_by: Union[_Missing, List[Author]] = missing
    parent: Union[_Missing, ModelParent] = missing
    run_mode: Union[_Missing, RunMode] = missing
    sample_inputs: Union[_Missing, List[URI]] = missing
    sample_outputs: Union[_Missing, List[URI]] = missing
    sha256: Union[_Missing, str] = missing
    source: Union[_Missing, ImportableSource] = missing
    test_inputs: List[URI] = missing
    test_outputs: List[URI] = missing
    timestamp: datetime = missing
    weights: Dict[WeightsFormat, WeightsEntry] = missing


# Manifest
BioImageIoManifest = dict
BioImageIoManifestModelEntry = dict
BioImageIoManifestNotebookEntry = dict
Badge = dict



import os
import pathlib

from ruamel.yaml import YAML

try:
    from typing import Literal, get_args
except ImportError:
    from typing_extensions import Literal, get_args  # noqa


yaml = YAML(typ="safe")

BIOIMAGEIO_CACHE_PATH = pathlib.Path(os.getenv("BIOIMAGEIO_CACHE_PATH", pathlib.Path.home() / "bioimageio_cache"))


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]



import datetime
import hashlib
import os
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import numpy as np

import bioimageio.spec as spec


#
# utility functions to build the spec from python
#


def _get_local_path(uri, root=None):
    is_local_path = os.path.exists(uri)
    if not is_local_path and root is not None:
        uri2 = os.path.join(root, uri)
        if os.path.exists(uri2):
            uri = uri2
            is_local_path = True
    if not is_local_path:
        uri = spec.fields.URI().deserialize(uri)
        uri = spec.download_uri_to_local_path(uri).as_posix()
    return uri


def _get_hash(path):
    with open(path, "rb") as f:
        data = f.read()
        return hashlib.sha256(data).hexdigest()


def _infer_weight_type(path):
    ext = os.path.splitext(path)[-1]
    if ext in (".pt", ".torch"):
        return "pytorch_state_dict"
    elif ext in (".pickle", ".pkl"):
        return "pickle"
    elif ext == ".onnx":
        return "onnx"
    elif ext in (".hdf", ".hdf5", ".h5"):
        return "keras_hdf5"
    elif ext == ".zip":
        return "tensorflow_saved_model_bundle"
    elif ext == ".json":
        return "tensorflow_js"
    else:
        raise ValueError(f"Could not infer weight type from extension {ext} for weight file {path}")


def _get_weights(weight_uri, weight_type, source, root, **kwargs):
    weight_path = _get_local_path(weight_uri, root)
    if weight_type is None:
        weight_type = _infer_weight_type(weight_path)
    weight_hash = _get_hash(weight_path)

    # if we have a "::" this is a python file with class specified,
    # so we can compute the hash for it
    if source is not None and "::" in source:
        source_path = _get_local_path(source.split("::")[0], root)
        source_hash = _get_hash(source_path)
    else:
        source_hash = None

    if "weight_attachments" in kwargs:
        attachments = {"attachments": ["weight_attachments"]}
    else:
        attachments = {}

    weight_types = spec.raw_nodes.WeightsFormat
    if weight_type == "pytorch_state_dict":
        # pytorch-state-dict -> we need a source
        assert source is not None
        weights = spec.raw_nodes.WeightsEntry(source=weight_uri, sha256=weight_hash, **attachments)
        language = "python"
        framework = "pytorch"

    elif weight_type == "pickle":
        weights = spec.raw_nodes.WeightsEntry(source=weight_uri, sha256=weight_hash, **attachments)
        language = "python"
        framework = "scikit-learn"

    elif weight_type == "onnx":
        weights = spec.raw_nodes.WeightsEntry(
            source=weight_uri, sha256=weight_hash, opset_version=kwargs.get("opset_version", 12), **attachments
        )
        language = None
        framework = None

    elif weight_type == "pytorch_script":
        weights = spec.raw_nodes.WeightsEntry(source=weight_uri, sha256=weight_hash, **attachments)
        if source is None:
            language = None
            framework = None
        else:
            language = "python"
            framework = "pytorch"

    elif weight_type == "keras_hdf5":
        weights = spec.raw_nodes.WeightsEntry(
            source=weight_uri,
            sha256=weight_hash,
            tensorflow_version=kwargs.get("tensorflow_version", "1.15"),
            **attachments,
        )
        language = "python"
        framework = "tensorflow"

    elif weight_type == "tensorflow_saved_model_bundle":
        weights = spec.raw_nodes.WeightsEntry(
            source=weight_uri,
            sha256=weight_hash,
            tensorflow_version=kwargs.get("tensorflow_version", "1.15"),
            **attachments,
        )
        language = "python"
        framework = "tensorflow"

    elif weight_type == "tensorflow_js":
        weights = spec.raw_nodes.WeightsEntry(source=weight_uri, sha256=weight_hash, **attachments)
        language = None
        framework = None

    elif weight_type in weight_types:
        raise ValueError(f"Weight type {weight_type} is not supported yet in 'build_spec'")
    else:
        raise ValueError(f"Invalid weight type {weight_type}, expect one of {weight_types}")

    weights = {weight_type: weights}
    return weights, language, framework, source_hash


def _get_data_range(data_range, dtype):
    if data_range is None:
        if np.issubdtype(np.dtype(dtype), np.integer):
            min_, max_ = np.iinfo(dtype).min, np.iinfo(dtype).max
        # for floating point numbers we assume valid range from -inf to inf
        elif np.issubdtype(np.dtype(dtype), np.floating):
            min_, max_ = -np.inf, np.inf
        elif np.issubdtype(np.dtype(dtype), np.bool):
            min_, max_ = 0, 1
        else:
            raise RuntimeError(f"Cannot derived data range for dtype {dtype}")
        data_range = (min_, max_)
    assert isinstance(data_range, (tuple, list))
    assert len(data_range) == 2
    return data_range


def _get_axes(axes, ndim):
    if axes is None:
        assert ndim in (2, 4, 5)
        default_axes = {2: "bc", 4: "bcyx", 5: "bczyx"}
        axes = default_axes[ndim]
    return axes


def _get_input_tensor(test_in, name, step, min_shape, data_range, axes, preprocessing):
    shape = test_in.shape
    if step is None:
        assert min_shape is None
        shape_description = shape
    else:
        shape_description = {"min": shape if min_shape is None else min_shape, "step": step}

    axes = _get_axes(axes, test_in.ndim)
    data_range = _get_data_range(data_range, test_in.dtype)

    kwargs = {}
    if preprocessing is not None:
        kwargs["preprocessing"] = preprocessing

    inputs = spec.raw_nodes.InputTensor(
        name="input" if name is None else name,
        data_type=str(test_in.dtype),
        axes=axes,
        shape=shape_description,
        data_range=data_range,
        **kwargs,
    )
    return inputs


def _get_output_tensor(test_out, name, reference_input, scale, offset, axes, data_range, postprocessing, halo):
    shape = test_out.shape
    if reference_input is None:
        assert scale is None
        assert offset is None
        shape_description = shape
    else:
        assert scale is not None
        assert offset is not None
        shape_description = {"reference_input": reference_input, "scale": scale, "offset": offset}

    axes = _get_axes(axes, test_out.ndim)
    data_range = _get_data_range(data_range, test_out.dtype)

    kwargs = {}
    if postprocessing is not None:
        kwargs["postprocessing"] = postprocessing
    if halo is not None:
        kwargs["halo"] = halo

    outputs = spec.raw_nodes.OutputTensor(
        name="output" if name is None else name,
        data_type=str(test_out.dtype),
        axes=axes,
        data_range=data_range,
        shape=shape_description,
        **kwargs,
    )
    return outputs


def _build_authors(authors: List[Dict[str, str]]):
    return [spec.raw_nodes.Author(**a) for a in authors]


# TODO The citation entry should be improved so that we can properly derive doi vs. url
def _build_cite(cite: Dict[str, str]):
    citation_list = [spec.raw_nodes.CiteEntry(text=k, url=v) for k, v in cite.items()]
    return citation_list


# TODO we should make the name more specific: "build_model_spec"?
# TODO maybe "build_raw_model" as it return raw_nodes.Model
# NOTE does not support multiple input / output tensors yet
# to implement this we should wait for 0.4.0, see also
# https://github.com/bioimage-io/spec-bioimage-io/issues/70#issuecomment-825737433
def build_spec(
    weight_uri: str,
    test_inputs: List[str],
    test_outputs: List[str],
    # general required
    name: str,
    description: str,
    authors: List[Dict[str, str]],
    tags: List[str],
    license: str,
    documentation: str,
    covers: List[str],
    cite: Dict[str, str],
    root: Optional[str] = None,
    # model specific optional
    source: Optional[str] = None,
    model_kwargs: Optional[Dict[str, Union[int, float, str]]] = None,
    weight_type: Optional[str] = None,
    sample_inputs: Optional[str] = None,
    sample_outputs: Optional[str] = None,
    # tensor specific
    input_name: Optional[str] = None,
    input_step: Optional[List[int]] = None,
    input_min_shape: Optional[List[int]] = None,
    input_axes: Optional[str] = None,
    input_data_range: Optional[List[Union[int, str]]] = None,
    output_name: Optional[str] = None,
    output_reference: Optional[str] = None,
    output_scale: Optional[List[int]] = None,
    output_offset: Optional[List[int]] = None,
    output_axes: Optional[str] = None,
    output_data_range: Optional[List[Union[int, str]]] = None,
    halo: Optional[List[int]] = None,
    preprocessing: Optional[List[Dict[str, Dict[str, Union[int, float, str]]]]] = None,
    postprocessing: Optional[List[Dict[str, Dict[str, Union[int, float, str]]]]] = None,
    # general optional
    git_repo: Optional[str] = None,
    attachments: Optional[Dict[str, Union[str, List[str]]]] = None,
    packaged_by: Optional[List[str]] = None,
    run_mode: Optional[str] = None,
    parent: Optional[str] = None,
    config: Optional[Dict[str, Any]] = None,
    dependencies: Optional[str] = None,
    **weight_kwargs,
):
    """Create a bioimageio.spec.schema.Model object that can be used to serialize a model.yaml in the bioimage.io format.

    Example usage:
    ```
    import bioimageio.spec as spec
    model_spec = spec.utils.build_spec(
        weight_uri="test_weights.pt",
        test_inputs=["./test_inputs"],
        test_outputs=["./test_outputs"],
        name="My Model",
        description="My very fancy model.",
        authors=[{"name": "John Doe", "affiliation": "My Institute"}],
        tags=["segmentation", "light sheet data"],
        license="CC-BY",
        documentation="./documentation.md",
        covers=["./my_cover.png"],
        cite={"Architecture": "https://my_architecture.com"}
    )
    serialized = spec.schema.Model().dump(model_spec)
    with open('model.yaml', 'w') as f:
        spec.utils.yaml.dump(serialized, f)
    ```

    Args:
        weight_uri: the url or relative local file path to the weight file for this model.
        test_inputs: list of test input files stored in numpy format.
        test_outputs: list of test outputs corresponding to test_inputs, stored in numpy format.
        name: name of this model.
        description: short description of this model.
        authors: the authors of this model.
        tags: list of tags for this model.
        license: the license for this model.
        documentation: relative file path to markdown documentation for this model.
        covers: list of relative file paths for cover images.
        cite: citations for this model.
        root: root folder for arguments given as relative paths.
        source: the file with the source code for the model architecture and the corresponding class.
        model_kwargs: the keyword arguments for the model class.
        weight_type: the type of the weights.
        sample_inputs: list of sample inputs to demonstrate the model performance.
        sample_outputs: list of sample outputs corresponding to sample_inputs.
        input_name: name of the input tensor.
        input_step: minimal valid increase of the input tensor shape.
        input_min_shape: minimal input tensor shape.
        input_axes: axes names for the input tensor.
        input_data_range: valid data range for the input tensor.
        output_name: name of the output tensor.
        output_reference: name of the input reference tensor used to cimpute the output tensor shape.
        output_scale: multiplicative factor to compute the output tensor shape.
        output_offset: additive term to compute the output tensor shape.
        output_axes: axes names of the output tensor.
        output_data_range: valid data range for the output tensor.
        halo: halo to be cropped from the output tensor.
        preprocessing: list of preprocessing operations for the input.
        postprocessing: list of postprocessing operations for the output.
        git_repo: reference git repository for this model.
        attachments: list of additional files to package with the model.
        packaged_by: list of authors that have packaged this model.
        run_mode: custom run mode for this model.
        parent: id of the parent mode from which this model is derived.
        config: custom configuration for this model.
        dependencies: relative path to file with dependencies for this model.
        weight_kwargs: keyword arguments for this weight type, e.g. "tensorflow_version".
    """
    #
    # generate the model specific fields
    #

    # check the test inputs and auto-generate input/output description from test inputs/outputs
    for test_in, test_out in zip(test_inputs, test_outputs):
        test_in, test_out = _get_local_path(test_in, root), _get_local_path(test_out, root)
        test_in, test_out = np.load(test_in), np.load(test_out)
    inputs = _get_input_tensor(
        test_in, input_name, input_step, input_min_shape, input_axes, input_data_range, preprocessing
    )
    outputs = _get_output_tensor(
        test_out,
        output_name,
        output_reference,
        output_scale,
        output_offset,
        output_axes,
        output_data_range,
        postprocessing,
        halo,
    )

    (weights, language, framework, source_hash) = _get_weights(weight_uri, weight_type, source, root, **weight_kwargs)

    #
    # generate general fields
    #
    format_version = spec.__version__
    timestamp = datetime.datetime.now()

    if source is not None:
        source = spec.fields.ImportableSource().deserialize(source)

    # optional kwargs, don't pass them if none
    optional_kwargs = {
        "git_repo": git_repo,
        "attachments": attachments,
        "packaged_by": packaged_by,
        "parent": parent,
        "run_mode": run_mode,
        "config": config,
        "sample_inputs": sample_inputs,
        "sample_outputs": sample_outputs,
        "framework": framework,
        "language": language,
        "source": source,
        "sha256": source_hash,
        "kwargs": model_kwargs,
        "dependencies": dependencies,
    }
    kwargs = {k: v for k, v in optional_kwargs.items() if v is not None}

    # build raw_nodes objects
    authors = _build_authors(authors)
    cite = _build_cite(cite)
    documentation = Path(documentation)
    covers = [spec.fields.URI().deserialize(uri) for uri in covers]
    test_inputs = [spec.fields.URI().deserialize(uri) for uri in test_inputs]
    test_outputs = [spec.fields.URI().deserialize(uri) for uri in test_outputs]

    model = spec.raw_nodes.Model(
        format_version=format_version,
        name=name,
        description=description,
        authors=authors,
        cite=cite,
        tags=tags,
        license=license,
        documentation=documentation,
        covers=covers,
        timestamp=timestamp,
        weights=weights,
        inputs=[inputs],
        outputs=[outputs],
        test_inputs=test_inputs,
        test_outputs=test_outputs,
        **kwargs,
    )

    # serialize and deserialize the raw_nodes.Model to
    # check that all fields are well formed
    serialized = spec.schema.Model().dump(model)
    model = spec.schema.Model().load(serialized)

    return model


def add_weights(model, weight_uri: str, root: Optional[str] = None, weight_type: Optional[str] = None, **weight_kwargs):
    """Add weight entry to bioimage.io model."""
    new_weights = _get_weights(weight_uri, weight_type, None, root, **weight_kwargs)[0]
    model.weights.update(new_weights)

    serialized = spec.schema.Model().dump(model)
    model = spec.schema.Model().load(serialized)

    return model


def serialize_spec(model, out_path):  # TODO change name to include model (see build_model_spec)
    serialized = spec.schema.Model().dump(model)
    spec.utils.yaml.dump(serialized, out_path)
